@@ -304,13 +304,13 @@ bool parse_caps(cap_bits_t *caps, char *names) {
 
   assert(8 * sizeof set >= (unsigned) max_cap);
 
-  while ((tok = strsep(&names, ","))) {
+  while (names && *names && (tok = strsep(&names, ","))) {
     rc = cap_from_name(tok, &val);
     if (rc != 0) {
       fprintf(stderr, "cannot interpret capability \"%s\"\n", tok);
       good = false;
     }
-    set |= (1 << val);
+    set |= (1ull << val);
   }
   *caps = set;
   return good;
