@@ -29,8 +29,13 @@ static int pivot_root(const char *new_root, const char *put_old) {
   return syscall(SYS_pivot_root, new_root, put_old);
 }
 
-static const char *special_mount_names[] = {
-  [SPECIAL_PROC] = "proc",
+static const struct {
+  char *name;
+  int ns;
+  char *fs;
+} special_mount_id[] = {
+  [SPECIAL_PROC] = { "proc", CLONE_NEWPID, "proc"  },
+  [SPECIAL_MAX]  = { NULL, 0, NULL }
 };
 
 static const char *exclude_root_dirs[] = {
@@ -47,55 +52,38 @@ static ssize_t num_mounts = 0;
 static bool bind_root_dirs(const char *new_root) {
   enum special_mount special;
   const struct dirent *de;
-  struct stat statbuf;
-  const char *entity;
   const char **filter;
   DIR *dir = NULL;
   char *path = NULL;
   struct mount_info *mt;
   bool success = false;
-  int dir1 = -1;
-  int dir2 = -1;
 
-  mts = NULL;
   memset(special_mounts, '\0', sizeof special_mounts);
 
-  if ((dir1 = open(entity = "/", O_RDONLY | O_DIRECTORY)) == -1) {
+  if ((dir = opendir("/")) == NULL) {
     perror("opening root directory");
     return false;
   }
-  if ((dir2 = openat(dir1, "./", O_RDONLY | O_DIRECTORY)) == -1)
-    goto fail;
-  if (fstat(dir1, &statbuf) == -1) {
-    perror("stat");
-    goto fail;
-  }
-  if ((dir = fdopendir(dir1)) == NULL) {
-    perror("opening root directory");
-    close(dir1);
-    return false;
-  }
 
-  mts = calloc(statbuf.st_nlink, sizeof *mts);
-  if (mts == NULL) {
-    perror("calloc");
-    goto fail;
-  }
-  mt = mts;
-  for (errno = 0; (de = readdir(dir));) {
+  for (num_mounts = 0; (de = readdir(dir)); num_mounts++);
+  rewinddir(dir);
+  mts = calloc(num_mounts + 1, sizeof *mts);
+
+  for (mt = mts; mt - mts < num_mounts && (de = readdir(dir)); mt++) {
     if (asprintf(&mt->from, "/%s", de->d_name) == -1 ||
         asprintf(&mt->to, "%s/%s", new_root, de->d_name) == -1)
-      mt->to = NULL;
+      goto fail;
 
     if (de->d_type == DT_LNK) {
       int rc;
       path = malloc(PATH_MAX);
       if (path == NULL)
         goto fail;
-      rc = readlinkat(dir2, de->d_name, path, PATH_MAX - 1);
+      rc = readlinkat(dirfd(dir), de->d_name, path, PATH_MAX - 1);
       if (rc == -1 || rc >= PATH_MAX - 1)
         goto fail;
       path[rc]='\0';
+      errno = 0;
       if (symlink(path, mt->to) == -1 || is_verbose())
         fprintf(stderr, "  symlink(%s,%s)=%s\n", path, mt->to, strerror(errno));
       free(path);
@@ -109,10 +97,10 @@ static bool bind_root_dirs(const char *new_root) {
     if (*filter)
       continue;
 
-    for (special = 0; special < SPECIAL_MAX && strcmp(de->d_name, special_mount_names[special]); special++);
+    for (special = 0; special < SPECIAL_MAX && strcmp(de->d_name, special_mount_id[special].name); special++);
 
     if (is_debug()) {
-      fprintf(stderr, "binding %s into new rootfs\n", de->d_name);
+      fprintf(stderr, "creating mount point %s in new rootfs\n", de->d_name);
     }
     mkdir(mt->to, 0700);
     if (special < SPECIAL_MAX) {
@@ -120,17 +108,20 @@ static bool bind_root_dirs(const char *new_root) {
         fprintf(stderr, "  found special mount %s\n", de->d_name);
       special_mounts[special] = mt;
     }
-    if (special != SPECIAL_PROC || (opt.new_ns & CLONE_NEWPID) == 0) {
-      mt->rc = mount(mt->from, mt->to, NULL, MS_BIND | MS_REC | MS_PRIVATE, NULL);
-      if (is_debug() || mt->rc != 0)
-        fprintf(stderr, "  mount(%s,%s)=%s\n", mt->from, mt->to, strerror(mt->rc == 0 ? 0 : errno));
-      if (mt->rc == 0)
-        mt->mounted = true;
+
+    if (opt.new_ns & special_mount_id[special].ns) {
+      free(mt->from);
+      mt->from = strdup(special_mount_id[special].name);
+      mt->rc = special_mount(mt->to, special_mount_id[special].fs, mt->from, NULL);
+    } else {
+      mt->rc = mount(mt->from, mt->to, NULL, MS_BIND | MS_REC | MS_SLAVE, NULL);
     }
 
-    mt++;
+    if (is_debug() || mt->rc != 0)
+      fprintf(stderr, "  mount(%s,%s)=%s\n", mt->from, mt->to, strerror(mt->rc == 0 ? 0 : errno));
+    if (mt->rc == 0)
+      mt->mounted = true;
   }
-  num_mounts = mt - mts;
 
   success = true;
 fail:
@@ -143,8 +134,6 @@ fail:
   }
   free(path);
   closedir(dir);
-  if (dir2 != -1)
-    close(dir2);
   return success;
 }
 
@@ -219,7 +208,7 @@ bool create_new_root(const char *executable,
     perror("special_mount: root");
     goto finish;
   }
-  rc = mount("", rt->path, "", MS_REMOUNT | MS_BIND | MS_REC | MS_PRIVATE, "");
+  rc = mount("", rt->path, "", MS_REMOUNT | MS_BIND | MS_REC | MS_SLAVE, "");
   if (rc == -1) {
     perror("adjusting new root fs mount options");
     goto finish;
@@ -237,11 +226,6 @@ bool pivot_to_new_root(struct pivot_state *rt) {
   bool success = false;
   int old_dir = -1;
   int rc;
-  char *pwd;
-
-  pwd = get_current_dir_name();
-  if (pwd == NULL)
-    perror("warning: could not save location of starting directory");
 
   if (chdir(rt->path) == -1)
     perror("chdir to new root");
@@ -268,7 +252,7 @@ bool pivot_to_new_root(struct pivot_state *rt) {
     perror("could not open old root to tidy up");
   }
 
-  if (mount("", ".", "", MS_SLAVE | MS_REC, NULL) == -1)
+  if (mount("", ".", "", MS_REC | MS_SLAVE, NULL) == -1)
     perror("could not disable old root mount propagation");
 
   if (umount2(".", MNT_DETACH) == -1)
@@ -280,6 +264,7 @@ bool pivot_to_new_root(struct pivot_state *rt) {
     success = true;
 
   close(old_dir);
+  old_dir = -1;
 
   rc = umount2(rt->path, MNT_DETACH);
   if (rc == -1)
@@ -289,9 +274,9 @@ bool pivot_to_new_root(struct pivot_state *rt) {
   if (rc == -1)
     perror("unlinkat");
 
-  if (pwd != NULL && chdir(pwd) == -1)
-    perror("warning: could not change to original directory relative to new root");
-
 finish:
+  if (old_dir != -1)
+    close(old_dir);
+
   return success;
 }
